/*
 * libmm-imgp-gstcs
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <stdbool.h>
#include "mm_util_gstcs_internal.h"
#include <gst/check/gstcheck.h>
#include <gst/video/video-format.h>

#define MM_UTIL_ROUND_UP_2(num) (((num)+1)&~1)
#define MM_UTIL_ROUND_UP_4(num) (((num)+3)&~3)
#define MM_UTIL_ROUND_UP_8(num) (((num)+7)&~7)
#define MM_UTIL_ROUND_UP_16(num) (((num)+15)&~15)

#define SAFE_STRCPY(dst, src, n)	g_strlcpy(dst, src, n)


static GstFlowReturn
_mm_sink_sample(GstElement * appsink, gpointer user_data)
{
	GstBuffer *_buf = NULL;
	GstSample *_sample = NULL;
	gstreamer_s * pGstreamer_s = (gstreamer_s*) user_data;
	_sample = gst_app_sink_pull_sample((GstAppSink*)appsink);
	if (_sample) {
		_buf = gst_sample_get_buffer(_sample);

		pGstreamer_s->output_buffer = _buf;

		if (pGstreamer_s->output_buffer != NULL) {
			GstMapInfo mapinfo = GST_MAP_INFO_INIT;
			gst_buffer_map(pGstreamer_s->output_buffer, &mapinfo, GST_MAP_READ);
			gstcs_debug("Create Output Buffer: GST_BUFFER_DATA: %p\t GST_BUFFER_SIZE: %zu", mapinfo.data, mapinfo.size);
			gst_buffer_unmap(pGstreamer_s->output_buffer, &mapinfo);
		} else {
			gstcs_error("ERROR -Input Prepare Buffer! Check createoutput buffer function");
		}
	}

	gst_buffer_ref(pGstreamer_s->output_buffer); /* when you want to avoid flushing */
	gst_sample_unref(_sample);

	return GST_FLOW_OK;
}

static gboolean
_mm_on_src_message(GstBus * bus, GstMessage * message, gpointer user_data)
{
	gstreamer_s * pGstreamer_s = (gstreamer_s*) user_data;
	switch (GST_MESSAGE_TYPE(message)) {
	case GST_MESSAGE_EOS: {
		gstcs_debug("The source got dry");
		gst_app_src_end_of_stream(GST_APP_SRC(pGstreamer_s->appsrc));
		g_main_context_pop_thread_default(pGstreamer_s->context);
		g_main_loop_quit(pGstreamer_s->loop);
		break;
		}
	case GST_MESSAGE_ERROR: {
		GError *err = NULL;
		gchar *dbg_info = NULL;

		gst_message_parse_error(message, &err, &dbg_info);
		gstcs_error("ERROR from element %s: %s\n", GST_OBJECT_NAME(message->src), err->message);
		gstcs_error("Debugging info: %s\n", (dbg_info) ? dbg_info : "none");
		g_error_free(err);
		g_free(dbg_info);
		g_main_context_pop_thread_default(pGstreamer_s->context);
		g_main_loop_quit(pGstreamer_s->loop);
		gstcs_debug("Quit GST_CS\n");
		break;
		}
	default:
		break;
	}
	return TRUE;
}

static int _mm_get_byte_per_pixcel(mm_util_color_format_e color_format)
{
	int byte_per_pixcel = 1;

	switch (color_format) {
	case MM_UTIL_COLOR_YUV420:
	case MM_UTIL_COLOR_YUV422:
	case MM_UTIL_COLOR_I420:
	case MM_UTIL_COLOR_NV12:
	case MM_UTIL_COLOR_UYVY:
	case MM_UTIL_COLOR_YUYV:
		byte_per_pixcel = 1;
		break;
	case MM_UTIL_COLOR_RGB16:
		byte_per_pixcel = 2;
		break;
	case MM_UTIL_COLOR_RGB24:
		byte_per_pixcel = 3;
		break;
	case MM_UTIL_COLOR_ARGB:
	case MM_UTIL_COLOR_BGRA:
	case MM_UTIL_COLOR_RGBA:
	case MM_UTIL_COLOR_BGRX:
		byte_per_pixcel = 4;
		break;
	default:
		gstcs_error("Not supported format");
		break;
	}

	gstcs_debug("color_format [%d] byte per pixcel [%d]", color_format, byte_per_pixcel);

	return byte_per_pixcel;
}

static int _mm_create_pipeline(gstreamer_s* pGstreamer_s)
{
	int ret = GSTCS_ERROR_NONE;
	pGstreamer_s->pipeline = gst_pipeline_new("pipeline");
	if (!pGstreamer_s->pipeline) {
		gstcs_error("pipeline could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	pGstreamer_s->appsrc = gst_element_factory_make("appsrc" , "appsrc");
	if (!pGstreamer_s->appsrc) {
		gstcs_error("appsrc could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	pGstreamer_s->colorspace = gst_element_factory_make("videoconvert" , "convert");
	if (!pGstreamer_s->colorspace) {
		gstcs_error("colorspace could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	pGstreamer_s->videoscale = gst_element_factory_make("videoscale", "scale");
	if (!pGstreamer_s->videoscale) {
		gstcs_error("videoscale could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	pGstreamer_s->videoflip = gst_element_factory_make("videoflip", "flip");
	if (!pGstreamer_s->videoflip) {
		gstcs_error("videoflip could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	pGstreamer_s->appsink = gst_element_factory_make("appsink" , "appsink");
	if (!pGstreamer_s->appsink) {
		gstcs_error("appsink could not be created. Exiting.\n");
		ret = GSTCS_ERROR_INVALID_PARAMETER;
	}
	return ret;
}

static void _mm_destroy_notify(gpointer data)
{
	unsigned char *_data = (unsigned char *)data;
	if (_data != NULL) {
		free(_data);
		_data = NULL;
	}
}

static void
_mm_check_caps_format(GstCaps* caps)
{
	GstStructure *caps_structure = gst_caps_get_structure(caps, 0);
	const gchar* formatInfo = gst_structure_get_string(caps_structure, "format");
	gstcs_debug("[%d] caps: %s", GST_IS_CAPS(caps), formatInfo);
}

static void
_mm_link_pipeline(gstreamer_s* pGstreamer_s, int value)
{
	gstcs_fenter();

	/* set property */
	gst_bin_add_many(GST_BIN(pGstreamer_s->pipeline), pGstreamer_s->appsrc, pGstreamer_s->colorspace, pGstreamer_s->videoscale, pGstreamer_s->videoflip, pGstreamer_s->appsink, NULL);
	if (!gst_element_link_many(pGstreamer_s->appsrc, pGstreamer_s->colorspace, pGstreamer_s->videoscale, pGstreamer_s->videoflip, pGstreamer_s->appsink, NULL))
		gstcs_error("Fail to link pipeline");
	else
		gstcs_debug("Success to link pipeline");

	g_object_set(G_OBJECT(pGstreamer_s->appsrc), "stream-type", 0, "format", GST_FORMAT_TIME, NULL);
	g_object_set(pGstreamer_s->appsrc, "num-buffers", 1, NULL);
	g_object_set(pGstreamer_s->appsrc, "is-live", TRUE, NULL); /* add because of gstreamer_s time issue */

	g_object_set(pGstreamer_s->videoflip, "method", value, NULL); /* GST_VIDEO_FLIP_METHOD_IDENTITY (0): none- Identity (no rotation) (1): clockwise - Rotate clockwise 90 degrees (2): rotate-180 - Rotate 180 degrees (3): counterclockwise - Rotate counter-clockwise 90 degrees (4): horizontal-flip - Flip horizontally (5): vertical-flip - Flip vertically (6): upper-left-diagonal - Flip across upper left/lower right diagonal (7): upper-right-diagonal - Flip across upper right/lower left diagonal */

	/*g_object_set(pGstreamer_s->appsink, "drop", TRUE, NULL);*/
	g_object_set(pGstreamer_s->appsink, "emit-signals", TRUE, "sync", FALSE, NULL);

	gstcs_fleave();
}

static GstVideoFormat _mm_get_video_format(mm_util_color_format_e color_format)
{
	GstVideoFormat videoFormat = GST_VIDEO_FORMAT_UNKNOWN;
	int _bpp = 0;
	int _depth = 0;
	int _red_mask = 0;
	int _green_mask = 0;
	int _blue_mask = 0;
	int _alpha_mask = 0;
	int _endianness = 0;

	switch (color_format) {
	case MM_UTIL_COLOR_YUV420:
		videoFormat = GST_VIDEO_FORMAT_YV12;
		break;
	case MM_UTIL_COLOR_YUV422:
		videoFormat = GST_VIDEO_FORMAT_Y42B;
		break;
	case MM_UTIL_COLOR_I420:
		videoFormat = GST_VIDEO_FORMAT_I420;
		break;
	case MM_UTIL_COLOR_NV12:
		videoFormat = GST_VIDEO_FORMAT_NV12;
		break;
	case MM_UTIL_COLOR_UYVY:
		videoFormat = GST_VIDEO_FORMAT_UYVY;
		break;
	case MM_UTIL_COLOR_YUYV:
		videoFormat = GST_VIDEO_FORMAT_YVYU;
		break;
	case MM_UTIL_COLOR_RGB16:
		_bpp = 16; _depth = 16; _red_mask = 63488; _green_mask = 2016; _blue_mask = 31; _endianness = 1234; _alpha_mask = 0;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	case MM_UTIL_COLOR_RGB24:
		_bpp = 24; _depth = 24; _red_mask = 16711680; _green_mask = 65280; _blue_mask = 255; _endianness = 4321; _alpha_mask = 0;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	case MM_UTIL_COLOR_ARGB:
		_bpp = 32; _depth = 32; _red_mask = 16711680; _green_mask = 65280; _blue_mask = 255; _alpha_mask = -16777216; _endianness = 4321;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	case MM_UTIL_COLOR_BGRA:
		_bpp = 32; _depth = 32; _red_mask = 65280; _green_mask = 16711680; _blue_mask = -16777216; _alpha_mask = 255; _endianness = 4321;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	case MM_UTIL_COLOR_RGBA:
		_bpp = 32; _depth = 32; _red_mask = -16777216; _green_mask = 16711680; _blue_mask = 65280; _alpha_mask = 255; _endianness = 4321;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	case MM_UTIL_COLOR_BGRX:
		_bpp = 32; _depth = 24; _red_mask = 65280; _green_mask = 16711680; _blue_mask = -16777216; _endianness = 4321; _alpha_mask = 0;
		videoFormat = gst_video_format_from_masks(_depth, _bpp, _endianness, _red_mask, _green_mask, _blue_mask, _alpha_mask);
		break;
	default:
		gstcs_error("Not supported format");
		break;
	}

	gstcs_debug("color_format [%d], Chosen video format [%s]", color_format, gst_video_format_to_string(videoFormat));

	return videoFormat;

}

static GstCaps* _mm_get_capabilities(mm_util_color_format_e color_format, unsigned int width, unsigned int height)
{
	GstCaps *caps = NULL;
	GstVideoFormat videoFormat = GST_VIDEO_FORMAT_UNKNOWN;

	videoFormat = _mm_get_video_format(color_format);

	caps = gst_caps_new_simple("video/x-raw",
			"format", G_TYPE_STRING, gst_video_format_to_string(videoFormat),
			"framerate", GST_TYPE_FRACTION, 25, 1,
			"pixel-aspect-ratio", GST_TYPE_FRACTION, 1, 1,
			"width", G_TYPE_INT, width,
			"height", G_TYPE_INT, height,
			"framerate", GST_TYPE_FRACTION, 1, 1,
			NULL);

	if (caps)
		_mm_check_caps_format(caps);
	else
		gstcs_error("caps is NULL");

	return caps;
}

static void _mm_get_round_up_width_height(mm_util_color_format_e color_format, unsigned int width, unsigned int height, unsigned int *stride, unsigned int *elevation)
{
	switch (color_format) {
	case MM_UTIL_COLOR_YUV420:
	case MM_UTIL_COLOR_YUV422:
	case MM_UTIL_COLOR_I420:
	case MM_UTIL_COLOR_NV12:
	case MM_UTIL_COLOR_UYVY:
	case MM_UTIL_COLOR_YUYV:
		*stride = width;
		*elevation = height;
		break;
	case MM_UTIL_COLOR_RGB16:
	case MM_UTIL_COLOR_RGB24:
		*stride = MM_UTIL_ROUND_UP_4(width);
		*elevation = MM_UTIL_ROUND_UP_2(height);
		break;
	case MM_UTIL_COLOR_ARGB:
	case MM_UTIL_COLOR_BGRA:
	case MM_UTIL_COLOR_RGBA:
	case MM_UTIL_COLOR_BGRX:
		*stride = width;
		*elevation = MM_UTIL_ROUND_UP_2(height);
		break;
	default:
		gstcs_error("Not supported format");
		break;
	}

	gstcs_debug("color_format[%d] width[%u] height[%u] stride[%u], elevation[%u]", color_format, width, height, *stride, *elevation);
}

static size_t _mm_setup_image_size(mm_util_color_format_e color_format, unsigned int width, unsigned int height)
{
	size_t size = 0;

	gstcs_debug("color_format [%d] width [%u] height [%u]", color_format, width, height);

	switch (color_format) {
	case MM_UTIL_COLOR_YUV420:
		size = (MM_UTIL_ROUND_UP_4(width) * MM_UTIL_ROUND_UP_2(height) + MM_UTIL_ROUND_UP_8(width) * MM_UTIL_ROUND_UP_2(height) / 2); /* width * height *1; */
		break;
	case MM_UTIL_COLOR_YUV422:
		size = (MM_UTIL_ROUND_UP_4(width) * height + MM_UTIL_ROUND_UP_8(width) * height); /*width * height *2; */
		break;
	case MM_UTIL_COLOR_I420:
		size = (MM_UTIL_ROUND_UP_4(width) * MM_UTIL_ROUND_UP_2(height) + MM_UTIL_ROUND_UP_8(width) * MM_UTIL_ROUND_UP_2(height) /2); /*width * height *1.5; */
		break;
	case MM_UTIL_COLOR_NV12:
		size = (MM_UTIL_ROUND_UP_4(width) * MM_UTIL_ROUND_UP_2(height) * 1.5); /* width * height *1.5; */
		break;
	case MM_UTIL_COLOR_UYVY:
		size = (MM_UTIL_ROUND_UP_2(width) * 2 * height); /* width * height *2; */
		break;
	case MM_UTIL_COLOR_YUYV:
		size = (MM_UTIL_ROUND_UP_2(width) * 2 * height); /* width * height *2; */
		break;
	case MM_UTIL_COLOR_RGB16:
		size = (MM_UTIL_ROUND_UP_4(width) * 2 * height); /* width * height *2; */
		break;
	case MM_UTIL_COLOR_RGB24:
		size = (MM_UTIL_ROUND_UP_4(width) * 3 * height); /* width * height *3; */
		break;
	case MM_UTIL_COLOR_ARGB:
	case MM_UTIL_COLOR_BGRA:
	case MM_UTIL_COLOR_RGBA:
	case MM_UTIL_COLOR_BGRX:
		size = width * height *4;
		break;
	default:
		gstcs_error("Not supported format");
		break;
	}


	gstcs_debug("Image size [%zu]", size);

	return size;
}

static int _mm_push_buffer_into_pipeline(imgp_info_s* pImgp_info, unsigned char *src, gstreamer_s * pGstreamer_s)
{
	int ret = GSTCS_ERROR_NONE;
	size_t data_size = 0;
	GstBuffer* gst_buf = NULL;

	gstcs_fenter();

	gstcs_retvm_if(pGstreamer_s->pipeline == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid pipeline");

	data_size = _mm_setup_image_size(pImgp_info->src_format, pImgp_info->src_width, pImgp_info->src_height);
	gst_buf = gst_buffer_new_wrapped_full(GST_MEMORY_FLAG_READONLY, src, data_size, 0, data_size, NULL, NULL);

	gstcs_retvm_if(gst_buf == NULL, GSTCS_ERROR_INVALID_OPERATION, "buffer is NULL");

	gst_app_src_push_buffer(GST_APP_SRC(pGstreamer_s->appsrc), gst_buf); /* push buffer to pipeline */

	gstcs_fleave();

	return ret;
}

static int _mm_push_buffer_into_pipeline_new(unsigned char *src, gstreamer_s * pGstreamer_s, mm_util_color_format_e color_format, unsigned int width, unsigned int height, unsigned int stride, unsigned int elevation)
{
	int ret = GSTCS_ERROR_NONE;
	GstBuffer *gst_buf = NULL;
	size_t src_size = 0;
	unsigned char *data = NULL;

	gstcs_fenter();

	gstcs_retvm_if(pGstreamer_s->pipeline == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid pipeline");
	gstcs_retvm_if((width == 0 || height == 0), GSTCS_ERROR_INVALID_PARAMETER, "Invalid width(%u) and height(%u)", width, height);

	src_size = _mm_setup_image_size(color_format, stride, elevation);

	int byte_per_pixcel = _mm_get_byte_per_pixcel(color_format);
	unsigned int src_row = width * byte_per_pixcel;
	unsigned int stride_row = stride * byte_per_pixcel;
	unsigned int i = 0, y = 0;

	data = (unsigned char *) malloc(src_size);
	gstcs_retvm_if(data == NULL, GSTCS_ERROR_OUT_OF_MEMORY, "data is NULL");

	for (y = 0; y < height; y++) {
		guint8 *pLine = (guint8 *) &(src[src_row * y]);
		for (i = 0; i < src_row; i++)
			data[y * stride_row + i] = pLine[i];
		guint8 stride_row_color = pLine[i - 1];
		for (i = src_row; i < stride_row; i++)
			data[y * stride_row + i] = stride_row_color;
	}
	for (y = height; y < elevation; y++) {
		for (i = 0; i < stride_row; i++)
			data[y * stride_row + i] = data[(y - 1) * stride_row + i];
	}
	gst_buf = gst_buffer_new_wrapped_full(GST_MEMORY_FLAG_READONLY, data, src_size, 0, src_size, data, _mm_destroy_notify);

	if (gst_buf == NULL) {
		gstcs_error("buffer is NULL\n");
		GSTCS_FREE(data);
		return GSTCS_ERROR_INVALID_PARAMETER;
	}

	gst_app_src_push_buffer(GST_APP_SRC(pGstreamer_s->appsrc), gst_buf); /* push buffer to pipeline */

	gstcs_fleave();

	return ret;
}

static int _mm_imgp_gstcs_processing(gstreamer_s* pGstreamer_s, unsigned char *src, unsigned char **dst, imgp_info_s* pImgp_info)
{
	GstBus *bus = NULL;
	GstStateChangeReturn ret_state;
	int ret = GSTCS_ERROR_NONE;
	GstCaps *src_caps = NULL;
	GstCaps *dst_caps = NULL;
	unsigned int src_stride = 0;
	unsigned int src_elevation = 0;

	gstcs_fenter();

	/*create pipeline*/
	ret = _mm_create_pipeline(pGstreamer_s);
	if (ret != GSTCS_ERROR_NONE)
		gstcs_error("ERROR - mm_create_pipeline ");

	/* Make appsink emit the "new-preroll" and "new-sample" signals. This option is by default disabled because signal emission is expensive and unneeded when the application prefers to operate in pull mode. */
	gst_app_sink_set_emit_signals((GstAppSink*)pGstreamer_s->appsink, TRUE);

	bus = gst_pipeline_get_bus(GST_PIPELINE(pGstreamer_s->pipeline));
	gst_bus_add_watch(bus, (GstBusFunc) _mm_on_src_message, pGstreamer_s);
	gst_object_unref(bus);

	_mm_get_round_up_width_height(pImgp_info->src_format, pImgp_info->src_width, pImgp_info->src_height, &src_stride, &src_elevation);
	_mm_get_round_up_width_height(pImgp_info->dst_format, pImgp_info->dst_width, pImgp_info->dst_height, &pImgp_info->output_stride, &pImgp_info->output_elevation);

	src_caps = _mm_get_capabilities(pImgp_info->src_format, src_stride, src_elevation);
	dst_caps = _mm_get_capabilities(pImgp_info->dst_format, pImgp_info->output_stride, pImgp_info->output_elevation);

	gst_app_src_set_caps(GST_APP_SRC(pGstreamer_s->appsrc), src_caps);
	gst_app_sink_set_caps(GST_APP_SINK(pGstreamer_s->appsink), dst_caps);

	if ((pImgp_info->src_width != src_stride) || (pImgp_info->src_height != src_elevation)) {
		ret = _mm_push_buffer_into_pipeline_new(src, pGstreamer_s, pImgp_info->src_format, pImgp_info->src_width, pImgp_info->src_height, src_stride, src_elevation);
	} else {
		ret = _mm_push_buffer_into_pipeline(pImgp_info, src, pGstreamer_s);
	}

	if (ret != GSTCS_ERROR_NONE) {
		gstcs_error("ERROR - mm_push_buffer_into_pipeline ");
		if (src_caps)
			gst_caps_unref(src_caps);
		if (dst_caps)
			gst_caps_unref(dst_caps);
		gst_object_unref(pGstreamer_s->pipeline);
		return ret;
	}

	/*link pipeline*/
	_mm_link_pipeline(pGstreamer_s, pImgp_info->angle);

	/* Conecting to the new-sample signal emited by the appsink*/
	gstcs_debug("Start G_CALLBACK(_mm_sink_sample)");
	g_signal_connect(pGstreamer_s->appsink, "new-sample", G_CALLBACK(_mm_sink_sample), pGstreamer_s);
	gstcs_debug("End G_CALLBACK(_mm_sink_sample)");

	/* GST_STATE_PLAYING*/
	gstcs_debug("Start GST_STATE_PLAYING");
	ret_state = gst_element_set_state(pGstreamer_s->pipeline, GST_STATE_PLAYING);
	gstcs_debug("End GST_STATE_PLAYING ret_state: %d", ret_state);

	/*g_main_loop_run*/
	gstcs_debug("g_main_loop_run");
	g_main_loop_run(pGstreamer_s->loop);

	gstcs_debug("Sucess GST_STATE_CHANGE");

	/*GST_STATE_NULL*/
	gst_element_set_state(pGstreamer_s->pipeline, GST_STATE_NULL);
	gstcs_debug("End GST_STATE_NULL");

	gstcs_debug("###pGstreamer_s->output_buffer### : %p", pGstreamer_s->output_buffer);

	ret_state = gst_element_get_state(pGstreamer_s->pipeline, NULL, NULL, 1*GST_SECOND);

	if (ret_state == GST_STATE_CHANGE_SUCCESS)
		gstcs_debug("GST_STATE_NULL ret_state = %d (GST_STATE_CHANGE_SUCCESS)\n", ret_state);
	else if (ret_state == GST_STATE_CHANGE_ASYNC)
		gstcs_debug("GST_STATE_NULL ret_state = %d (GST_STATE_CHANGE_ASYNC)\n", ret_state);

	gstcs_debug("Success gst_element_get_state\n");

	if (src_caps)
		gst_caps_unref(src_caps);
	if (dst_caps)
		gst_caps_unref(dst_caps);

	if (ret_state == GST_STATE_CHANGE_FAILURE) {
		gstcs_error("GST_STATE_CHANGE_FAILURE");
	} else {
		if (pGstreamer_s->output_buffer != NULL) {
			GstMapInfo mapinfo = GST_MAP_INFO_INIT;
			gst_buffer_map(pGstreamer_s->output_buffer, &mapinfo, GST_MAP_READ);
			size_t buffer_size = mapinfo.size;
			size_t calc_buffer_size = 0;

			calc_buffer_size = _mm_setup_image_size(pImgp_info->dst_format, pImgp_info->output_stride, pImgp_info->output_elevation);

			gstcs_debug("buffer size[%zu], calc[%zu]", buffer_size, calc_buffer_size);
			if (buffer_size != calc_buffer_size) {
				gstcs_debug("Buffer size is different \n");
				gstcs_debug("unref output buffer");
				gst_buffer_unref(pGstreamer_s->output_buffer);
				gst_object_unref(pGstreamer_s->pipeline);
				pGstreamer_s->output_buffer = NULL;
				return GSTCS_ERROR_INVALID_OPERATION;
			}
			gstcs_debug("pGstreamer_s->output_buffer: %p\n", pGstreamer_s->output_buffer);
			*dst = calloc(1, buffer_size);
			if (*dst == NULL) {
				gstcs_error("ERROR - calloc ");
				gst_buffer_unref(pGstreamer_s->output_buffer);
				gst_object_unref(pGstreamer_s->pipeline);
				pGstreamer_s->output_buffer = NULL;
				return GSTCS_ERROR_INVALID_OPERATION;
			}

			memcpy(*dst, mapinfo.data, buffer_size);
			pImgp_info->buffer_size = buffer_size;
			gst_buffer_unmap(pGstreamer_s->output_buffer, &mapinfo);
		} else {
			gstcs_debug("pGstreamer_s->output_buffer is NULL");
		}
	}
	gstcs_debug("unref output buffer");
	gst_buffer_unref(pGstreamer_s->output_buffer);
	gst_object_unref(pGstreamer_s->pipeline);
	pGstreamer_s->output_buffer = NULL;

	gstcs_debug("dst: %p", *dst);
	gstcs_fleave();

	return ret;
}

static int _gstcs_create_default_thread(gstreamer_s *gstreamer)
{
	gstcs_retvm_if(gstreamer == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid gstreamer");

	gstreamer->context = g_main_context_new();
	gstcs_retvm_if(gstreamer->context == NULL, GSTCS_ERROR_INVALID_OPERATION, "ERROR - g_main_context_new");

	gstreamer->loop = g_main_loop_new(gstreamer->context, FALSE);
	if (gstreamer->loop == NULL) {
		gstcs_error("ERROR - g_main_loop_new ");
		g_main_context_unref(gstreamer->context);
		return GSTCS_ERROR_INVALID_OPERATION;
	}

	g_main_context_push_thread_default(gstreamer->context);

	return GSTCS_ERROR_NONE;
}

static int _gstcs_destroy_default_thread(gstreamer_s *gstreamer)
{
	gstcs_retvm_if(gstreamer == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid gstreamer");

	if (gstreamer->loop != NULL)
		g_main_loop_unref(gstreamer->loop);

	if (gstreamer->context != NULL)
		g_main_context_unref(gstreamer->context);

	return GSTCS_ERROR_NONE;
}

static int _gstcs_init(gstreamer_s** gstreamer)
{
	static const int max_argc = 50;
	gint argc = 0;
	gchar** argv = NULL;
	int i = 0;
	int ret = GSTCS_ERROR_NONE;

	argv = malloc(sizeof(gchar*) * max_argc);

	gstcs_retvm_if(argv == NULL, GSTCS_ERROR_OUT_OF_MEMORY, "argv is not allocated");

	memset(argv, 0, sizeof(gchar*) * max_argc);

	argv[argc] = g_strdup("mmutil_gstcs");
	if (argv[argc] == NULL) {
		gstcs_error("argv[%d] is not allocated", argc);
		ret = GSTCS_ERROR_OUT_OF_MEMORY;
	}
	argc++;
	/* check disable registry scan */
	argv[argc] = g_strdup("--gst-disable-registry-update");
	if (argv[argc] == NULL) {
		gstcs_error("argv[%d] is not allocated", argc);
		ret = GSTCS_ERROR_OUT_OF_MEMORY;
	}
	argc++;
	if (ret != GSTCS_ERROR_NONE) {
		for (i = 0; i < argc; i++)
			GSTCS_FREE(argv[i]);

		GSTCS_FREE(argv);
		return ret;
	}

	gst_init(&argc, &argv);

	*gstreamer = g_new0(gstreamer_s, 1);
	if (*gstreamer == NULL) {
		gstcs_error("gstreamer structure is not allocated");
		ret = GSTCS_ERROR_OUT_OF_MEMORY;
	}

	for (i = 0; i < argc; i++)
		GSTCS_FREE(argv[i]);

	GSTCS_FREE(argv);
	return ret;
}

static int _mm_imgp_gstcs(imgp_info_s* pImgp_info, unsigned char *src, unsigned char **dst)
{
	gstreamer_s* pGstreamer_s;
	int ret = GSTCS_ERROR_NONE;

	/* Print debug message for inout structure */
	gstcs_debug("src_width [%d] src_height [%d] src_format [%d] dst_width [%d] dst_height [%d] dst_format [%d] rotation [%d]",
	pImgp_info->src_width, pImgp_info->src_height, pImgp_info->src_format, pImgp_info->dst_width, pImgp_info->dst_height, pImgp_info->dst_format, pImgp_info->angle);

	/* Initialize gstreamer */
	ret = _gstcs_init(&pGstreamer_s);
	if (ret != GSTCS_ERROR_NONE) {
		gstcs_error("Error: _gstcs_new is failed");
		return ret;
	}

	/* Create default thread for async behavior */
	ret = _gstcs_create_default_thread(pGstreamer_s);
	if (ret != GSTCS_ERROR_NONE) {
		gstcs_error("Error: _gstcs_create_default_thread is failed");
		GSTCS_FREE(pGstreamer_s);
		return ret;
	}

	/* Do gstreamer processing */
	ret = _mm_imgp_gstcs_processing(pGstreamer_s, src, dst, pImgp_info); /* input: buffer pointer for input image , input image format, input image width, input image height, output: buffer porinter for output image */

	if (ret == GSTCS_ERROR_NONE)
		gstcs_debug("End _mm_imgp_gstcs_processing [dst: %p]", *dst);
	else if (ret != GSTCS_ERROR_NONE)
		gstcs_error("ERROR - _mm_imgp_gstcs_processing");

	/* Free resouces */
	ret = _gstcs_destroy_default_thread(pGstreamer_s);
	if (ret != GSTCS_ERROR_NONE)
		gstcs_error("Error: _gstcs_create_default_thread is failed");

	GSTCS_FREE(pGstreamer_s);

	return ret;
}

static bool _mm_imgp_check_format(mm_util_color_format_e color_format)
{
	if ((color_format >= MM_UTIL_COLOR_YUV420) && (color_format <= MM_UTIL_COLOR_BGRX))
		return TRUE;

	return FALSE;
}

int mm_imgp(imgp_info_s* pImgp_info, unsigned char *src, unsigned char **dst, imgp_type_e _imgp_type)
{
	gstcs_retvm_if(pImgp_info == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid info");
	gstcs_retvm_if(src == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid src");
	gstcs_retvm_if(dst == NULL, GSTCS_ERROR_INVALID_PARAMETER, "Invalid dst");
	gstcs_retvm_if((_imgp_type < 0 || _imgp_type > IMGP_MAX), GSTCS_ERROR_INVALID_PARAMETER, "Invalid _imgp_type[%d]", _imgp_type);
	gstcs_retvm_if((_mm_imgp_check_format(pImgp_info->src_format) == FALSE), GSTCS_ERROR_NOT_SUPPORTED_FORMAT, "not supported src_format [%d]", pImgp_info->src_format);
	gstcs_retvm_if((_mm_imgp_check_format(pImgp_info->dst_format) == FALSE), GSTCS_ERROR_NOT_SUPPORTED_FORMAT, "not supported dst_format [%d]", pImgp_info->dst_format);

	gstcs_debug("[src %p] [dst %p]", src, dst);

	return _mm_imgp_gstcs(pImgp_info, src, dst);
}
